using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using Upsilon.BaseTypes.UserData;
using Upsilon.Evaluator;

namespace Upsilon.BaseTypes.ScriptFunction
{
    internal class ScriptMethodInfoFunction : ScriptFunction
    {
        public ScriptMethodInfoFunction(UserDataMethod method, object o, bool directTypeManipulation,
            bool passScriptReference = false, bool passScopeReference = false)
        {
            _method                 = method;
            _object                 = o;
            _directTypeManipulation = directTypeManipulation;
            _passScriptReference = passScriptReference;
            _passScopeReference = passScopeReference;

            ReturnType              = _method.ReturnType;
        }

        private readonly UserDataMethod _method;
        private readonly object         _object;
        private readonly bool           _directTypeManipulation;
        private readonly bool           _passScriptReference;
        private readonly bool _passScopeReference;
        public           System.Type    ReturnType { get; }

        public IEnumerable<UserDataMethod.UserDataMethodParameter> GetParameterInfo()
        {
            return _method.GetMethods().First().Parameters;
        }

        public override ScriptType Run(Diagnostics diagnostics, ScriptType[] variables, Script script, EvaluationScope scope)
        {
            var types = _directTypeManipulation
                ? variables.Select(x => x.GetType()).ToArray()
                : variables.Select(x => x.GetCSharpType()).ToArray();
            var method = _method.GetMethod(types);
            if (method == null)
            {
                throw new Exception(
                    $"No valid function found on type '{_object.GetType()}' with name '{_method.Name}' " +
                    $"and parameter types: {string.Join(", ", types.Select(x => $"'{x.Name}'"))}");
            }
            var objects = new List<object>();
            if (_passScriptReference)
                objects.Add(script);
            if (_passScopeReference)
                objects.Add(scope);
            objects.AddRange(_directTypeManipulation
                ? variables.Select(x => (object) x).ToList()
                : variables.Select(x => x.ToCSharpObject()).ToList());
            var pars = method.GetParameters();
            if (pars.Length != objects.Count)
            {
                for (var i = objects.Count; i < pars.Length; i++)
                {
                    objects.Add(null);
                }
            }

            object result;
            try
            {
                result = method.Invoke(_object, objects.ToArray());
            }
            catch (TargetInvocationException e)
            {
                if (e.InnerException != null) throw e.InnerException;
                throw;
            }
            if (_directTypeManipulation)
                return (ScriptType)result;
            return result.ToScriptType();

        }
    }
}