using System;
using System.Collections.Generic;
using Upsilon.BaseTypes;
using Upsilon.BaseTypes.Number;
using Upsilon.BaseTypes.ScriptTypeInterfaces;
using Upsilon.Binder;
using Upsilon.Evaluator;

// ReSharper disable UnusedMember.Global
// ReSharper disable MemberCanBePrivate.Global

namespace Upsilon.StandardLibraries
{
    internal class BasicFunctions : ScriptLibrary
    {
        [StandardLibraryScriptFunction("assert", "Asserts that the parameter passed is true. Throws an exception if it is not true. " +
                                                 "Can take a message to show in the exception, otherwise throws with message \"assertion failed!\"")]
        public void Assert(ScriptBoolean boolean, ScriptString message = null)
        {
            if (!boolean)
            {
                Error(message ?? new ScriptString("assertion failed!"));
            }
        }

        [StandardLibraryScriptFunction("error", "Throw an exception with given error message")]
        public void Error(ScriptString message)
        {
            throw new Exception(message.Value);
        }

        [StandardLibraryScriptFunction("ipairs", "Iterates over an iterable variable, like a table, until it encounters a nil value.")]
        public IIterable UpTillNullPairs(IIterable table)
        {
            return new UpTillNullPairsScriptIterator(table);
        }

        [StandardLibraryScriptFunction("pairs", "Iterates over an iterable variable, like a table, skipping all nil values.")]
        public IIterable Pairs(IIterable table)
        {
            return new PairsScriptIterator(table);
        }

        [StandardLibraryScriptFunction("require", "Loads a module from the module path, using the given script loader.",
            true)]
        public ScriptType Require(Script script, ScriptString fileName)
        {
            var file = fileName.Value;
            var module = script.Options.ModuleHandler.GetModule(script, file);
            var s = Executor.ContinueWith(script, module);
            var val = s.Evaluate();
            return val != null ? val.ToScriptType() : new ScriptNull();
        }


        [StandardLibraryScriptFunction("tonumber", "Parses a string to a number.")]
        public ScriptNumber ToNumber(ScriptString obj)
        {
            var str = obj.Value;
            if (str.Contains("."))
            {
                return new ScriptNumberDouble(double.Parse(str));
            }
            else
            {
                return new ScriptNumberLong(long.Parse(str));
            }
        }

        [StandardLibraryScriptFunction("tostring", "Returns the string value of the given object.")]
        public ScriptString ToString(ScriptType obj)
        {
            return new ScriptString(obj.ToString());
        }

        [StandardLibraryScriptFunction("type", "Returns the type of the given object as a string.")]
        public ScriptString Type(ScriptType obj)
        {
            return new ScriptString(obj.Type.ToString());
        }

    }
}